# (C) Copyright 2024 Anemoi contributors.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
#
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


from typing import Any
from typing import Dict
from typing import Protocol


class MetadataProtocol(Protocol):
    """Protocol for metadata objects. This will keep `mypy` happy."""

    _metadata: Dict[str, Any]
    _supporting_arrays: Dict[str, Any]
    grid: str
    variables: Dict[str, Any]
